import functools
import cPickle
import jsonpickle

import minsol


class Acceptor(minsol.Object):
    """
    .. note::
        first argument ``callback`` should be ignored, if your
        client is not asynchronous.
    """

    def get_method(self, name):
        try:
            method = getattr(self, name)
            a = method.is_rpc_method
            del a

        except AttributeError:
            raise MethodNotFound()

        return method


class Request(minsol.Object):
    possible_keys = ["id", "method", "args", "kwargs"]

    def init(self, ident, method_name, args, kwargs, **other):
        self.id = ident
        self.method_name = method_name
        self.args = args
        self.kwargs = kwargs

    def execute(self, acceptor, callback):
        method = acceptor.get_method(self.method_name)

        try:
            method(callback, *self.args, **self.kwargs)

        except TypeError as error:
            raise InvalidParams(data=error.message)

    def as_raw(self):
        request = {
            "id": self.id,
            "method": self.method_name
        }

        if self.args:
            request["args"] = self.args

        if self.kwargs:
            request["kwargs"] = self.kwargs

        raw = self.serialize(request)
        return raw

    @classmethod
    def get_from_raw(cls, raw):
        try:
            raw = str(raw)
            request = cls.unserialize(raw)

        except ValueError as error:
            raise ParseError(data=error.message)

        if request.__class__ != dict:
            raise InvalidRequest("Request MUST be dict!")

        for k in request.keys():
            if k not in cls.possible_keys:
                raise InvalidRequest("Bad key %s" % k)

        if "method" not in request.keys():
            raise InvalidRequest("key 'method' MUST be specified!")

        if "args" in request.keys():
            if request["args"].__class__ != list:
                raise InvalidRequest("key 'args' MUST be list!")

        if "kwargs" in request.keys():
            if request["kwargs"].__class__ != dict:
                raise InvalidRequest("key 'kwargs' MUST be dict!")

        request_id = request.get("id", None)
        request = cls(
            ident=request_id, method_name=request["method"],
            args=request.get("args", []),
            kwargs=request.get("kwargs", {}))

        return request

    def serialize(self, request):
        raise NotImplementedError

    @classmethod
    def unserialize(cls, raw):
        raise NotImplementedError


class Response(minsol.Object):
    possible_keys = ["id", "result"]

    def init(self, request, ident=None, result=None, **kwargs):
        self.request = request
        self.raw_id = ident

        self.result = result

        if self.raw_id and self.request.id != self.raw_id:
            raise InvalidResponse()

    def as_raw(self):
        request_id = self.request.id if self.request else None

        response = {
            "id": request_id,
            "result": self.result
        }

        raw = self.serialize(response)
        return raw

    @classmethod
    def get_from_raw(cls, request, raw):
        try:
            raw = str(raw)
            response = cls.unserialize(raw)

        except ValueError as error:
            raise ParseError(data=error.message)

        if response.__class__ != dict:
            raise InvalidRequest("Response MUST be dict!")

        for k in response.keys():
            if k not in cls.possible_keys:
                raise InvalidRequest("Bad key %s" % k)

        if "result" not in response.keys():
            raise InvalidRequest("key 'result' MUST be specified!")

        response = cls(request=request, **response)
        return response

    def serialize(self, request):
        raise NotImplementedError

    @classmethod
    def unserialize(cls, raw):
        raise NotImplementedError


class Client(minsol.Object):
    request_class = Request
    response_class = Response

    def init(self, **kwargs):
        self.last_id = 1

    def __getattr__(self, name):
        if name[:2] == "__":
            raise AttributeError

        return functools.partial(self._rpc_call, name)

    def _rpc_call(self, method_name, *args, **kwargs):
        args = list(args)

        request = self.request_class(
            ident=self.last_id, method_name=method_name,
            args=args, kwargs=kwargs)

        self.last_id += 1

        raw_request = request.as_raw()
        raw_response = self._fetch_call(raw_request)

        response = self.response_class.get_from_raw(request, raw_response)

        result = response.result
        if isinstance(result, Exception):
            raise result

        return result

    def _fetch_call(self, raw_request):
        """
        :return: raw_request
        :rtype: str
        """
        raise NotImplementedError


class AsyncClient(Client):
    def _rpc_call(self, method_name, callback, *args, **kwargs):
        args = list(args)

        request = self.request_class(
            ident=self.last_id, method_name=method_name,
            args=args, kwargs=kwargs)

        self.last_id += 1

        raw_request = request.as_raw()

        def _fetched(raw_response):
            response = self.response_class.get_from_raw(request, raw_response)

            result = response.result
            if isinstance(result, Exception):
                callback(None, result)
                return

            callback(result, None)

        self._fetch_call(_fetched, raw_request)

    def _fetch_call(self, callback, raw_request):
        """
        :param callback: ``def callback(result, error)``
        :rtype: function
        """
        raise NotImplementedError


class JSONRequest(Request):
    def serialize(self, request):
        return jsonpickle.encode(request)

    @classmethod
    def unserialize(cls, raw):
        return jsonpickle.decode(raw)


class JSONResponse(Response):
    def serialize(self, response):
        return jsonpickle.encode(response)

    @classmethod
    def unserialize(cls, raw):
        return jsonpickle.decode(raw)


class JSONClient(Client):
    request_class = JSONRequest
    response_class = JSONResponse


class cPickleRequest(Request):
    def serialize(self, request):
        return cPickle.dumps(request)

    @classmethod
    def unserialize(cls, raw):
        return cPickle.loads(raw)


class cPickleResponse(Response):
    def serialize(self, request):
        return cPickle.dumps(request)

    @classmethod
    def unserialize(cls, raw):
        return cPickle.loads(raw)


class cPickleClient(Client):
    request_class = cPickleRequest
    response_class = cPickleResponse


class Error(minsol.Error):
    code = -32603
    message = 'Internal error'

    def __init__(self, data=None, request_id=None):
        self.data = data
        self.request_id = request_id

    def get_message(self):
        _str = minsol.Error.get_message(self)
        if self.data:
            _str += "\tDATA: "
            _str += self.data

        return _str


class ParseError(Error):
    code = -32700
    message = 'Parse error'


class InvalidRequest(Error):
    code = -32600
    message = 'Invalid request'


class MethodNotFound(Error):
    code = -32601
    message = 'Method not found'


class InvalidParams(Error):
    code = -32602
    message = 'Invalid params'


class InvalidResponse(Error):
    code = -32604
    message = 'Invalid response'


class ServerError(Error):
    code = -32000
    message = 'Server error'


def method(func):
    func.is_rpc_method = True
    return func


